def parse(uri):
    if "+" in uri:
        vcs, versioned_repo_uri = uri.split("+", 1)
    else:
        vcs = None
        versioned_repo_uri = uri
        
    if "#" in versioned_repo_uri:
        repo_uri, revision = versioned_repo_uri.split("#", 1)
    else:
        revision = None
        repo_uri = versioned_repo_uri
        
    return ParsedUri(vcs, repo_uri, revision)

class ParsedUri(object):
    def __init__(self, vcs, repo_uri, revision):
        self.vcs = vcs
        self.repo_uri = repo_uri
        self.revision = revision



import subprocess
import errno

_dev_null = open('/dev/null', 'w')

def quiet_check_call(command, cwd=None):
    return run(command, cwd=cwd).return_code

def quiet_call(command, cwd=None):
    return run(command, cwd=cwd, allow_error=True).return_code
    
def quiet_check_output(command, cwd):
    return run(command, cwd=cwd).output
    
def _quiet_subprocess_eval(func, args, kwargs):
    return func(*args, stdout=_dev_null, stderr=subprocess.STDOUT, **kwargs)

def run(command, cwd=None, allow_error=False):
    try:
        process = subprocess.Popen(command, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as error:
        raise NoSuchCommandError(command[0])
        
    output, stderr_output = process.communicate()
    return_code = process.poll()
    result = ExecutionResult(return_code, output)
    if not allow_error and return_code:
        error = subprocess.CalledProcessError(return_code, command)
        error.output = output
        error.stderr_output = stderr_output
        raise RuntimeError("error: " + stderr_output)
    return result

class ExecutionResult(object):
    def __init__(self, return_code, output):
        self.return_code = return_code
        self.output = output

class NoSuchCommandError(OSError):
    def __init__(self, command):
        if "/" in command:
            message = "No such command: {0}".format(command)
        else:
            message = "Command not found: {0}. Check that {0} is installed and on $PATH".format(command)
        super(type(self), self).__init__(message)
        self.command = command



import os

from blah import systems
from blah.files import parent

def repository_at(working_directory):
    directory = os.path.abspath(working_directory)
    files = os.listdir(directory)
    for system in systems.all_systems:
        if system.directory_name in files:
            return system.local_repo(directory)
        
    return None
    
def find_repository(directory):
    directory = os.path.abspath(directory)
    while directory is not None:
        repository = repository_at(directory)
        if repository is not None:
            return repository
        directory = parent(directory)
        
    return None



import os
import errno
import shutil
from contextlib import contextmanager
import tempfile

def mkdir_p(path):
    """ 'mkdir -p' in Python """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise
            
@contextmanager
def temporary_directory():
    directory = tempfile.mkdtemp()
    yield directory
    shutil.rmtree(directory)

def write_file(path, contents):
    with open(path, "w") as f:
        return f.write(contents)

def read_file(path):
    with open(path, "r") as f:
        return f.read()

def parent(file_path):
    parent = os.path.dirname(file_path)
    if file_path == parent:
        return None
    else:
        return parent



import os
import argparse

import blah.repositories
import blah.fetcher
import blah.errors

class WhatIsThisCommand(object):
    def create_parser(self, subparser):
        subparser.add_argument("directory", nargs="?")
    
    def execute(self, args):
        directory = args.directory if args.directory is not None else os.getcwd()
        repository = blah.repositories.find_repository(directory)
        if repository is None:
            print "Could not find source control repository"
        else:
            print "{0}+file://{1}".format(repository.type, repository.working_directory)

class FetchCommand(object):
    def create_parser(self, subparser):
        subparser.add_argument("repository_uri", metavar="repository-uri")
        subparser.add_argument("local_path", metavar="local-path")
        subparser.add_argument("--use-cache", default=False, help=argparse.SUPPRESS, action="store_true")
    
    def execute(self, args):
        try:
            blah.fetcher.fetch(args.repository_uri, args.local_path, args.use_cache)
        except (blah.errors.BlahUserError, blah.util.NoSuchCommandError) as error:
            print "fetch failed: {0}".format(error.message)
            exit(-1)

class ArchiveCommand(object):
    def create_parser(self, subparser):
        subparser.add_argument("repository_uri", metavar="repository-uri")
        subparser.add_argument("local_path", metavar="local-path")
    
    def execute(self, args):
        try:
            blah.fetcher.archive(args.repository_uri, args.local_path)
        except (blah.errors.BlahUserError, blah.util.NoSuchCommandError) as error:
            print "archive failed: {0}".format(error.message)
            exit(-1)


commands = {
    "what-is-this": WhatIsThisCommand(),
    "fetch": FetchCommand(),
    "archive": ArchiveCommand()
}



from blah.util import quiet_check_call, quiet_check_output

class Hg(object):
    name = "hg"
    directory_name = ".hg"
    default_branch = "default"
    
    def clone(self, repository_uri, local_path):
        quiet_check_call(["hg", "clone", repository_uri, local_path])
        return HgRepository(local_path)
        
    def local_repo(self, working_directory):
        return HgRepository(working_directory)
    

class HgRepository(object):
    type = Hg.name
    
    def __init__(self, working_directory):
        self.working_directory = working_directory
    
    def update(self):
        quiet_check_call(["hg", "pull"], cwd=self.working_directory)
        
    def checkout_revision(self, revision):
        quiet_check_call(["hg", "update", revision], cwd=self.working_directory)
        
    def remote_repo_uri(self):
        uri = quiet_check_output(["hg", "showconfig", "paths.default"], cwd=self.working_directory).strip()
        # Mercurial strips off "file:///" from the URI
        if _is_local(uri):
            return "file://" + uri
        else:
            return uri
            
    def head_revision(self):
        return quiet_check_output(["hg", "parents", "--template={node}"], cwd=self.working_directory).strip()
        
    def is_fixed_revision(self, revision):
        return False
        

def _is_local(uri):
    return "://" not in uri



from blah.fetcher import fetch, archive
from blah.repositories import find_repository, repository_at
from blah.errors import UnrecognisedSourceControlSystem
from blah.systems import is_source_control_uri

__all__ = [
    "fetch",
    "archive",
    "find_repository",
    "repository_at",
    "is_source_control_uri",
    "UnrecognisedSourceControlSystem"
]



from blah.git import Git
from blah.hg import Hg
import blah.uri_parser

all_systems = [
    Git(),
    Hg()
]

def is_source_control_uri(uri):
    parsed_uri = blah.uri_parser.parse(uri)
    return parsed_uri.vcs is not None and _is_recognised_prefix(parsed_uri.vcs)

def _is_recognised_prefix(prefix):
    return any(prefix == vcs.name for vcs in all_systems)



#!/usr/bin/env python

import os
from distutils.core import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='blah',
    version='0.1.11',
    description='Thin wrapper around source control systems',
    long_description=read("README"),
    author='Michael Williamson',
    url='http://github.com/mwilliamson/blah',
    scripts=["scripts/blah"],
    packages=['blah'],
    install_requires=["argparse==1.2.1", "catchy>=0.1.0"],
)



import os
import os.path
import hashlib

from blah.util import run
import blah.caching


class Git(object):
    name = "git"
    directory_name = ".git"
    default_branch = "origin/master"
    supports_caching = True
    
    def __init__(self, use_cache=False):
        self._use_cache = use_cache
        
    def use_cache(self):
        return Git(use_cache=True)
    
    def clone(self, repository_uri, local_path):
        if self._use_cache:
            cache_dir = self._update_cache(repository_uri)
            _git(["clone", repository_uri, local_path, "--reference", cache_dir])
            _git(["repack", "-a"], cwd=local_path)
            _git(["submodule", "foreach", "git", "repack", "-a"], cwd=local_path)
            
            os.remove(os.path.join(local_path, ".git/objects/info/alternates"))
        else:
            _git(["clone", repository_uri, local_path])
        return GitRepository(local_path)
        
    def local_repo(self, working_directory):
        return GitRepository(working_directory)
        
    def _update_cache(self, repository_uri):
        cache_root = blah.caching.cache_root()
        repo_hash = hashlib.sha1(repository_uri).hexdigest()
        cache_dir = os.path.join(cache_root, repo_hash)
        
        if os.path.exists(cache_dir):
            _git(["fetch"], cwd=cache_dir)
        else:
            _git(["clone", repository_uri, cache_dir])
            
        return cache_dir

class GitRepository(object):
    type = Git.name
    
    def __init__(self, working_directory):
        self.working_directory = working_directory
    
    def update(self):
        _git(["fetch"], cwd=self.working_directory)

    def checkout_revision(self, revision):
        if _git(["branch", "-r", "--contains", "origin/" + revision], cwd=self.working_directory, allow_error=True).return_code == 0:
            revision = "origin/" + revision
            
        _git(["checkout", revision], cwd=self.working_directory)

    def remote_repo_uri(self):
        return _git(["config", "remote.origin.url"], cwd=self.working_directory).output.strip()
        
    def head_revision(self):
        return _git(["rev-parse", "HEAD"], cwd=self.working_directory).output.strip()
        
    def is_fixed_revision(self, revision):
        if revision is None:
            return False
        else:
            return self._is_tag(revision)
        
    def _is_tag(self, revision):
        return bool(_git(["tag", "--list", revision], cwd=self.working_directory).output.strip())

def _git(git_command, *args, **kwargs):
    command = ["git"] + git_command
    return run(command, *args, **kwargs)



import os
import shutil
import hashlib

import catchy

import blah.systems
import blah.uri_parser
import blah.errors
import blah.caching

def archive(uri_str, local_path):
    uri_hash = _sha1(uri_str)
    cacher = catchy.DirectoryCacher(os.path.join(blah.caching.cache_root(), "archive"))
    cache_result = cacher.fetch(uri_hash, local_path)
    if not cache_result.cache_hit:
        uri = blah.uri_parser.parse(uri_str)
        
        vcs, local_repo = _fetch(uri_str, local_path)
        is_fixed_revision = local_repo.is_fixed_revision(uri.revision)
        shutil.rmtree(os.path.join(local_path, vcs.directory_name))
        
        if is_fixed_revision:
            cacher.put(uri_hash, local_path)

def _sha1(value):
    return hashlib.sha1(value).hexdigest()

# Define fetch as distinct from _fetch to stop return value leaking
def fetch(*args, **kwargs):
    _fetch(*args, **kwargs)

def _fetch(uri_str, local_path, use_cache=False, systems=None):
    if systems is None:
        systems = blah.systems.all_systems
    
    uri = blah.uri_parser.parse(uri_str)
    
    vcs = _find_vcs(uri, systems)
    if use_cache and getattr(vcs, "supports_caching", False):
        vcs = vcs.use_cache()
    local_repo = _fetch_all_revisions(uri, local_path, vcs)
    revision = _read_revision(vcs, uri)
    local_repo.checkout_revision(revision)
    return vcs, local_repo

def _find_vcs(uri, systems):
    for vcs in systems:
        if uri.vcs == vcs.name:
            return vcs
            
    message = "Source control system not recognised: {0}".format(uri.vcs)
    raise blah.errors.UnrecognisedSourceControlSystem(message)

def _fetch_all_revisions(uri, local_path, vcs):
    if os.path.exists(local_path):
        return _update(uri.repo_uri, local_path, vcs)
    else:
        return vcs.clone(uri.repo_uri, local_path)

def _read_revision(vcs, uri):
    if uri.revision is None:
        return vcs.default_branch
    else:
        return uri.revision

def _update(repository_uri, local_path, vcs):
    vcs_directory = os.path.join(local_path, vcs.directory_name)
    if not os.path.isdir(local_path):
        message = "Checkout path already exists, and is not directory: {0}".format(local_path)
        raise blah.errors.BlahUserError(message)
    elif not os.path.isdir(vcs_directory):
        message = "{0} already exists and is not a {1} repository".format(local_path, vcs.name)
        raise blah.errors.BlahUserError(message)
    else:
        local_repo = vcs.local_repo(local_path)
        current_remote_uri = local_repo.remote_repo_uri()
        if current_remote_uri == repository_uri:
            local_repo.update()
            return local_repo
        else:
            message = "{0} is existing checkout of different repository: {1}" \
                .format(local_path, current_remote_uri)
            raise blah.errors.BlahUserError(message)
    
